# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide the main function to be used as an entry point for the CLI."""


import argparse
import logging
import sys
from getpass import getpass
from typing import Optional
from urllib.error import HTTPError

from . import __version__
from .api import get_latest_tag


logger = logging.getLogger("tag_spy")


def parse_arguments():
    """Parse all command line arguments."""
    parser = argparse.ArgumentParser(description="Find the latest Docker image tag.")
    parser.add_argument(
        "image",
        metavar="IMAGE",
        help="A Docker image specification, for example, dddecaf/wsgi-base.",
    )
    parser.add_argument(
        "tag",
        metavar="BASE_TAG",
        help="The first part of an image tag. So if your tags have the format "
        "'<image>:<base>_<date>_<commit>', you should supply the <base> here.",
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--verbosity",
        help="The desired log level (default WARNING).",
        choices=("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
        default="WARNING",
    )
    parser.add_argument(
        "--authentication",
        metavar="URL",
        help="The URL from where to retrieve an access token for the registry "
        "(the default https://auth.docker.io/token corresponds to Docker Hub).",
        default="https://auth.docker.io/token",
    )
    parser.add_argument(
        "--registry",
        metavar="URL",
        help="The URL of the container registry API "
        "(default https://registry-1.docker.io).",
        default="https://registry-1.docker.io",
    )
    parser.add_argument(
        "--service",
        metavar="URI",
        help="The service URI for which to request an access token "
        "(default registry.docker.io).",
        default="registry.docker.io",
    )
    parser.add_argument(
        "--username",
        metavar="STR",
        help="A username used for authentication when retrieving an access token. "
        "Without a username no authentication will be attempted.",
        default=None,
    )
    parser.add_argument(
        "--password",
        action="store_true",
        help="A password will be requested interactively. It is used for "
        "authentication when retrieving an access token.The password field may "
        "itself be a token for Bearer authentication when the username is "
        "'oauth2accesstoken'.",
    )
    parser.add_argument(
        "--password-stdin",
        action="store_true",
        help="A password will be read from standard input. It is used for "
        "authentication when retrieving an access token.The password field may "
        "itself be a token for Bearer authentication when the username is "
        "'oauth2accesstoken'.",
    )
    args = parser.parse_args()
    logging.basicConfig(level=args.verbosity, format="[%(levelname)s] %(message)s")
    if args.password_stdin:
        password: Optional[str] = sys.stdin.read().rstrip("\r\n")
    else:
        password = getpass() if args.password else None
    args.password = password
    return args


def main() -> None:
    """Define the console script entry point and thus the command line interface."""
    args = parse_arguments()
    try:
        print(
            get_latest_tag(
                image=args.image,
                base_tag=args.tag,
                authentication_url=args.authentication,
                registry_url=args.registry,
                service=args.service,
                username=args.username,
                password=args.password,
            )
        )
    except HTTPError as error:
        logger.debug("", exc_info=error)
        logger.debug("%r", error.geturl())
        logger.debug("%d %s", error.status, error.reason)  # type: ignore
        logger.debug("HTTP Headers\n%s", error.info())
        logger.critical(str(error))



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide an HTTP client for making requests."""


import gzip
import json
import logging
import zlib
from http.client import HTTPResponse
from typing import Dict, List, Optional, Union
from urllib.parse import urlencode, urljoin, urlsplit, urlunsplit
from urllib.request import OpenerDirector, Request


logger = logging.getLogger(__name__)


class Response:
    """
    Define an HTTP response class.

    This class should not have to be instantiated directly.

    Attributes:
        status (int): The numeric HTTP status code of the response.
        body (bytes): The raw body of the response as bytes.

    Methods:
        json: Deserialize the response's body as JSON and return the object.

    """

    def __init__(self, *, response: HTTPResponse, **kwargs) -> None:
        """
        Initialize an HTTP response object.

        Args:
            response (http.client.HTTPResponse): A standard library response that is
                wrapped by this object.
            **kwargs: Passed to parent classes.
        """
        super().__init__(**kwargs)  # type: ignore
        self._response = response
        # Immediately read the body while the response context is still available.
        self._body = self._content()

    def _content(self) -> bytes:
        """Return the response body and decompress it if required."""
        if self._response.getheader("Content-Encoding", "").lower() == "gzip":
            return gzip.decompress(self._response.read())
        elif self._response.getheader("Content-Encoding", "").lower() == "deflate":
            return zlib.decompress(self._response.read())
        else:
            return self._response.read()

    @property
    def status(self) -> int:
        """Get the response's HTTP status code."""
        return self._response.status

    @property
    def body(self) -> bytes:
        """Get the response's raw body content as bytes."""
        return self._body

    def json(self) -> Union[Dict, List, str]:
        """Deserialize the response's body as JSON and return the object."""
        if len(self.body) == 0:
            return ""
        try:
            return json.loads(self.body)  # type: ignore
        except json.JSONDecodeError:
            logger.debug(self.body)
            raise


class Client:
    """
    Define an HTTP client for making requests.

    Methods:
        get: Make an HTTP GET request and return the response.

    """

    def __init__(
        self,
        *,
        opener: OpenerDirector,
        base_url: str,
        token: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initialize an HTTP client with a base URL.

        Args:
            opener (urllib.request.OpenerDirector): The opener with attached handlers
                and headers for calling the URL.
            base_url (str): The base URL for all future requests.
            token (str, optional): A bearer access token.
            **kwargs: Passed to parent constructors.

        """
        super().__init__(**kwargs)  # type: ignore
        self._opener = opener
        self._parts = urlsplit(base_url)
        self.token = token

    def _build_request(
        self, path: Optional[str] = None, params: Optional[Dict[str, str]] = None,
    ) -> Request:
        """Construct a correctly encoded URL from its parts."""
        url = self._parts
        if path is not None:
            url = url._replace(path=urljoin(url.path, path))
        if params is not None:
            url = url._replace(query=urlencode(params))
        request = Request(urlunsplit(url))
        if self.token:
            request.add_unredirected_header("Authorization", f"Bearer {self.token}")
        return request

    def get(
        self,
        path: Optional[str] = None,
        params: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Response:
        """
        Make an HTTP GET request and return the response.

        Args:
            path (str, optional): An optional relative or absolute URL path to modify
                the base URL.
            params (dict, optional): An optional mapping from query parameter names to
                their values.

        Returns:
            Response: An HTTP response instance.

        """

        logger.debug("Making GET request to %r.", self._build_request(path, params))
        with self._opener.open(self._build_request(path, params), **kwargs) as response:
            logger.debug("%r", response.geturl())
            logger.debug("%d %s", response.status, response.reason)
            logger.debug("HTTP Headers\n%s", response.info())
            # Must be returned within the context such that the response body is
            # available for reading.
            return Response(response=response)



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide helpers that interact with a Docker registry."""


import logging
import re
from datetime import datetime, timezone
from operator import itemgetter
from typing import List

from .http_client import Client
from .image_tag_triple import ImageTagTriple


logger = logging.getLogger(__name__)


def parse_iso_8601_timestamp(timestamp: str) -> datetime:
    """Parse a timestamp in ISO 8601 format to a datetime object."""

    def shorten_to_milliseconds(match: re.Match) -> str:
        return f"{match.group(1)}{match.group(2)[:6]}{match.group(3)}"

    return datetime.strptime(
        re.sub(r"([.])(\d*)([Z+-].*)$", shorten_to_milliseconds, timestamp),
        "%Y-%m-%dT%H:%M:%S.%f%z",
    )


def get_token(client: Client, image: str, service: str) -> str:
    """
    Return an access token for the requested image and service.

    Args:
        client (Client): A client instance pointed at the authentication endpoint.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        service (str): The URL of the service for which to request an access token.

    Returns:
        str: The access token which is valid for a pre-specified amount of time.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    logger.debug("Retrieving access token for image %r and service %r.", image, service)
    data = client.get(
        params={"scope": f"repository:{image}:pull", "service": service}
    ).json()
    assert isinstance(data, dict)
    return str(data["token"])


def verify_v2_capability(client: Client) -> None:
    """
    Verify that the registry API actually supports version 2.

    Args:
        client (Client): A client instance authenticated with a Bearer token and pointed
            at version 2 of the Docker registry API.

    Raises:
        urllib.error.HTTPError: In case the API does *not* support version 2.

    """
    logger.debug("Verifying version 2 API capability.")
    # The following statement will raise an HTTPError if the API does not support
    # version 2.
    # FIXME: Could be handled more nicely by raising an informative error.
    assert client.get().status == 200


def get_tags(client: Client, image: str) -> List[str]:
    """
    Return the list of tags for an image in its registry.

    Args:
        client (Client): A client instance authenticated with a Bearer token and pointed
            at version 2 of the Docker registry API.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.

    Returns:
        list: All the tags as strings that were found.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    logger.debug("Retrieving image %r tags.", image)
    data = client.get(path=f"{image}/tags/list").json()
    assert isinstance(data, dict)
    return [str(t) for t in data["tags"]]


def get_image_digest(client: Client, image: str, tag: str) -> str:
    """
    Return an image's digest from its manifest.

    Args:
        client (Client): A client instance authenticated with a Bearer token and pointed
            at version 2 of the Docker registry API.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        tag (str): The base part of the tag that you are interested in, for
            example, 'alpine' will match 'dddecaf/wsgi-base:alpine_2020-04-28_24fe0a0'.

    Returns:
        str: The image's digest hash.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    logger.info("Retrieving image '%s:%s' digest.", image, tag)
    data = client.get(path=f"{image}/manifests/{tag}").json()
    assert isinstance(data, dict)
    return str(data["config"]["digest"])


def get_image_creation_timestamp(client: Client, image: str, digest: str,) -> datetime:
    """
    Return an image's creation timestamp.

    Args:
        client (Client): A client instance authenticated with a Bearer token and pointed
            at version 2 of the Docker registry API.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        digest (str): An image digest as can be retrieved from its manifest.

    Returns:
        datetime.datetime: The timestamp that records when the image was built.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    See Also:
        get_image_digest

    """
    logger.info("Retrieving image %r configuration.", image)
    data = client.get(path=f"{image}/blobs/{digest}").json()
    assert isinstance(data, dict)
    try:
        return parse_iso_8601_timestamp(data["created"])
    except KeyError:
        logger.error("The 'created' key does not exist. Ignoring this image.")
        # Set the timestamp to the beginning of the epoch so it will never be the
        # latest.
        return datetime.fromtimestamp(0, timezone.utc)


def get_latest_by_timestamp(
    client: Client, image: str, tags: List[ImageTagTriple],
) -> ImageTagTriple:
    """
    Return the latest image tag as determined by the build timestamp.

    Args:
        client (Client): A client instance authenticated with a Bearer token and pointed
            at version 2 of the Docker registry API.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        tags (list): A collection of ``ImageTagTriple`` that all contain the same date.

    Returns:
        ImageTagTriple: The latest of the collection.

    Raises:
        urllib.error.URLError: In case of problems communicating with the registry.

    """
    latest = []
    for triple in tags:
        digest = get_image_digest(client, image, str(triple))
        logger.debug("Digest: %s", digest)
        build_timestamp = get_image_creation_timestamp(client, image, digest)
        logger.debug("        %s", build_timestamp)
        latest.append((triple, build_timestamp))
    latest.sort(key=itemgetter(1), reverse=True)
    return latest[0][0]



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide factory functions for creating opener directors."""


from typing import List, Tuple
from urllib.request import (
    HTTPBasicAuthHandler,
    HTTPPasswordMgrWithDefaultRealm,
    OpenerDirector,
    build_opener,
)


def generate_standard_headers() -> List[Tuple[str, str]]:
    """Generate a standard list of HTTP header, value pairs for openers."""
    return [
        ("Accept", "application/vnd.docker.distribution.manifest.v2+json"),
        ("Accept-Encoding", "gzip, deflate"),
    ]


def opener_factory() -> OpenerDirector:
    """Return a basic opener director with standard headers."""
    opener = build_opener()
    opener.addheaders.extend(generate_standard_headers())
    return opener


def basic_authentication_opener_factory(
    registry_url: str, username: str, password: str,
) -> OpenerDirector:
    """
    Return an opener director that authenticates requests using the Basic scheme.

    Args:
        registry_url (str): The URL of the container registry API.
        username (str): A username used for Basic authentication when retrieving an
            access token.
        password (str): A password used for Basic authentication when
            retrieving an access token.

    Returns:
        OpenerDirector: A director for making HTTP requests. It's main method is `open`.

    """
    opener = opener_factory()
    password_manager = HTTPPasswordMgrWithDefaultRealm()
    password_manager.add_password(
        None,  # type: ignore
        registry_url,
        username,
        password,
    )
    opener.add_handler(HTTPBasicAuthHandler(password_manager))
    return opener



# Copyright 2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Provide helpers that interact with a Docker registry."""


import logging
from typing import Optional
from urllib.parse import urljoin

from .http_client import Client
from .opener_factory import basic_authentication_opener_factory, opener_factory
from .registry_helpers import get_token


logger = logging.getLogger(__name__)


def authenticated_client_factory(
    registry_url: str,
    authentication_url: str,
    image: str,
    service: str,
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> Client:
    """
    Create a client instance which is authenticated with an access token.

    Args:
        registry_url (str): The URL of the container registry API.
        authentication_url (str): The URL from where to retrieve an access
            token for the registry.
        image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
        service (str): The service URI for which to request an access token.
        username (str, optional): A username used for authentication when
            retrieving an access token. Without a username no authentication will be
            attempted which will only succeed with public registries.
        password (str, optional): A password used for authentication when
            retrieving an access token. The password field may itself be a token for
            Bearer authentication iff the username is 'oauth2accesstoken'.

    Returns:
        Client: An HTTP client equipped with a Bearer token according to the privileges
            granted by the provided credentials.

    """
    if username is not None:
        assert password is not None, "Authentication requires a password or token."
        if username.lower() == "oauth2accesstoken":
            logger.debug("Retrieving access token using Bearer authentication.")
            token = get_token(
                Client(
                    opener=opener_factory(), base_url=authentication_url, token=password
                ),
                image,
                service,
            )
        else:
            logger.debug("Retrieving access token using Basic authentication.")
            token = get_token(
                Client(
                    opener=basic_authentication_opener_factory(
                        registry_url, username, password
                    ),
                    base_url=authentication_url,
                ),
                image,
                service,
            )
    else:
        logger.debug("Attempting to retrieve access token without credentials.")
        token = get_token(
            Client(opener=opener_factory(), base_url=authentication_url), image, service
        )
    return Client(
        opener=opener_factory(),
        # This package is built on version 2 of the Docker registry API.
        base_url=urljoin(registry_url, "/v2/"),
        token=token,
    )
